import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pdb

filename = 'final_grades.csv' # 'test_cases.csv' # 

# Read the input CSV
df = pd.read_csv(filename, header=None)
df = df.replace('-', 0) # replace dashes '-' with zero marks
# replace the column name 'Graded Quiz' with the 'Programming Assignment'
# for convenience in computing the aggregate score
df = df.replace('C1M3 Graded Quiz (Real)',
                'C1M3 Programming Assignment (Real)__')
dat = df.to_numpy()
num_students = len(dat) - 1

# column number which corresponds to the CCID
CCID_idx = np.where(dat[0] == 'CCID')[0].item()

# list of all the course components (excluding the Capstone)
# see the Syllabus (https://szepi.github.io/cmput-365-w23/syllabus.html)
course_components = ['Practice Quiz',
                     'Programming Assignment',
                     'WA1',
                     'WA2',
                     'Midterm total',
                     'Final total']

# total (maximum achievable) marks for each of the course components
total_marks = {
    'Practice Quiz': 10 * 100,
    'Programming Assignment': 9 * 100,
    'WA1': 100, 
    'WA2': 100, 
    'Midterm total': 50, 
    'Final total': 95
}

# the weightage of each of the course components
course_weightage = {
    'Practice Quiz': 0.1,
    'Programming Assignment': 0.3,
    'WA1': 0.1, 
    'WA2': 0.1, 
    'Midterm total': 0.15, 
    'Final total': 0.25
}

def find_idx(arr, substring):
    '''
    This function returns the column indices of 'arr' which have the 
    'substring' in the column name.
    '''
    return np.where(np.char.find(arr, substring) != -1)[0]

def create_dict():
    '''
    Create an empty dictionary which contains the marks of the students.
    '''
    student_marks = {
        'Practice Quiz': None,
        'Programming Assignment': None,
        'WA1': None, 
        'WA2': None, 
        'Midterm total': None, 
        'Final total': None
    }
    return student_marks

def find_best_course_component(percentages_dict, weight_remaining):
    '''
    Find the course element of the student (excluding the final) which 
    results in the maximum increase in the student's aggregated score, 
    if the percent score obtained on that course component were to 
    be increased to the percentage obtained on the final.
    '''
    final_per = percentages_dict['Final total']
    max_val = 0
    max_key = None
    for key in percentages_dict.keys(): # go through each key in the dictionary
        # current_val computes how much the student's aggregate score
        # will increase if this course component were to be increased
        # to become same as the final
        current_val = (final_per - tmp_dict[key]) \
            * min(weight_remaining, course_weightage[key])
        if current_val > max_val: # find the key which has the max value
            max_val = current_val
            max_key = key
    return max_key

# store the aggregate score achieved by the student
aggregate_score_list = []

for i in range(1, num_students+1): # iterate through all the students
    
    # create an empty dictionary that stores the component-wise marks
    marks = create_dict() 

    #===========================================================================
    # compute the marks for the individual course components
    #---------------------------------------------------------------------------
    for comp in course_components:
        marks[comp] = dat[i][
            find_idx(dat[0].astype(str), comp)].astype(float).sum()
    #===========================================================================

    #===========================================================================
    # compute marks for Practice Quizzes
    #---------------------------------------------------------------------------
    practice_quiz_list = dat[i][find_idx(
        dat[0].astype(str), 'Practice Quiz')].astype(float)
    practice_quiz_list.sort() # sort in an ascending order
    # pick the top 10 practice quizzes (out of a total of 11)
    marks['Practice Quiz'] = practice_quiz_list[1:].sum()
    #===========================================================================

    #===========================================================================
    # compute marks for Programming Assignments and use Capstone to replace
    # the bottom two assignments
    #---------------------------------------------------------------------------
    prog_assgn_list = dat[i][find_idx(
        dat[0].astype(str), 'Programming Assignment')].astype(float)
    prog_assgn_list.sort() # sort in an ascending order
    # pick the top 9 practice quizzes (out of a total of 10)
    prog_assgn_list = prog_assgn_list[1:]
    capstone = dat[i][find_idx(dat[0].astype(str), 'Capstone')].astype(float)
    if prog_assgn_list[0] < capstone:
        prog_assgn_list[0] = capstone
    if prog_assgn_list[1] < capstone:
        prog_assgn_list[1] = capstone
    marks['Programming Assignment'] = prog_assgn_list.sum()
    #===========================================================================

    #===========================================================================
    # Use the final exam's score to replace upto 15% of the course components
    # (optimize this replacement so that the components with maximum possible
    #  gain in marks are updated first)
    #---------------------------------------------------------------------------
    percent_score = {} # compute the percentage for each of the components
    for comp in course_components:
        percent_score[comp] = marks[comp] / total_marks[comp]
    
    # find the course components that when replaced by the grade in the final 
    # exam would give the highest increase in the aggregate percentage
    weight_remaining = 0.15
    tmp_dict = percent_score.copy()
    while weight_remaining > 0:
        comp = find_best_course_component(tmp_dict, weight_remaining)
        if comp is None: # no course component's percentage can be increased
            break
        tmp_dict.pop(comp)
        
        tmp_weight = min(weight_remaining, course_weightage[comp])
        weight_remaining = weight_remaining - tmp_weight
        replacement_weight = tmp_weight / course_weightage[comp]
        
        percent_score[comp] = \
            replacement_weight * percent_score['Final total'] \
            + (1 - replacement_weight) * percent_score[comp]
    #===========================================================================

    #===========================================================================
    # Compute the aggregate score (out of 100)
    #---------------------------------------------------------------------------
    aggregate_score = 100 * sum([percent_score[comp] * course_weightage[comp]
                                 for comp in course_components])
    #===========================================================================

    # add to a list
    aggregate_score_list.append([dat[i][CCID_idx], aggregate_score]) 
    
# save the list of aggregate scores as a CSV    
df2 = pd.DataFrame(aggregate_score_list, columns=['CCID', 'Aggregate Score'])
df2.to_csv('aggregate_scores.csv', index=False)

# plot a histogram for the aggregate scores
aggregate_score_arr = np.array(aggregate_score_list)[:, -1].astype(float)
plt.hist(aggregate_score_arr, bins=20, edgecolor='black', linewidth=1.2)
plt.xlabel('Marks')
plt.ylabel('# of students')
plt.title('Aggregate marks distribution\n'\
          'Min: {:.1f} | Max: {:.1f} | Average: {:.1f} | Median: {:.1f}'.format(
              np.min(aggregate_score_arr),
              np.max(aggregate_score_arr),
              np.mean(aggregate_score_arr),
              np.median(aggregate_score_arr)))
plt.savefig('aggregate_marks_distribution.png', dpi=300)
